// This file is part of Dust Racing 2D.
// Copyright (C) 2012 Jussi Lind <jussi.lind@iki.fi>
//
// Dust Racing 2D is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// Dust Racing 2D is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Dust Racing 2D. If not, see <http://www.gnu.org/licenses/>.

#include "statemachine.hpp"

#include "intro.hpp"
#include "race.hpp"
#include "renderer.hpp"
#include "startlights.hpp"
#include "track.hpp"

#include <MenuManager>

#include <cassert>

StateMachine * StateMachine::m_instance = nullptr;

static const float FADE_SPEED_MENU = 0.02;
static const float FADE_SPEED_GAME = 0.01;

StateMachine::StateMachine()
: m_state(Init)
, m_isFading(false)
, m_intro(nullptr)
, m_startlights(nullptr)
, m_race(nullptr)
, m_renderer(nullptr)
, m_track(nullptr)
, m_fadeValue(0.0)
, m_returnToMenu(false)
{
    assert(!StateMachine::m_instance);
    StateMachine::m_instance = this;
}

StateMachine::~StateMachine()
{
    StateMachine::m_instance = nullptr;
}

StateMachine & StateMachine::instance()
{
    assert(StateMachine::m_instance);
    return *StateMachine::m_instance;
}

bool StateMachine::update()
{
    switch (m_state)
    {
    case Init:

        m_state     = DoIntro;
        m_fadeValue = 0.0;

        break;

    case DoIntro:

        assert(m_renderer);

        if (m_intro->update()) // Intro return true when done.
        {
            m_renderer->setFadeValue(1.0);
            m_state = Menu;
        }

        break;

    case Menu:

        m_returnToMenu = false;
        m_fadeValue    = 1.0;
        m_isFading     = true; // Must be set to apply the initial fade value is set in Scene.
        m_renderer->setFadeValue(m_fadeValue);

        if (MTFH::MenuManager::instance().done())
        {
            m_state = MenuTransitionOut;
        }

        break;

    case MenuTransitionOut:

        if (m_fadeValue <= 0.0)
        {
            m_state     = GameTransitionIn;
            m_fadeValue = 0.0;
            m_isFading  = false;
        }
        else
        {
            m_fadeValue -= FADE_SPEED_MENU;
            m_isFading   = true;
        }

        m_renderer->setFadeValue(m_fadeValue);

        break;

    case GameTransitionIn:

        assert(m_track);

        if (m_fadeValue >= 1.0)
        {
            m_fadeValue = 1.0;
            m_isFading  = false;
            m_state     = DoStartlights;

            m_startlights->reset();
        }
        else
        {
            m_fadeValue += FADE_SPEED_MENU;
            m_isFading   = true;
        }

        m_renderer->setFadeValue(m_fadeValue);

        break;

    case DoStartlights:

        assert(m_startlights);

        if (!m_startlights->update())
        {
            m_state = Play;
        }

        break;

    case Play:

        assert(m_race);

        if (m_race->finished())
        {
            m_fadeValue = 3.0; // Add some delay with a value greater than 1.0
            m_state     = GameTransitionOut;
        }
        else if (m_returnToMenu)
        {
            m_fadeValue = 1.0;
            m_state     = GameTransitionOut;
        }

        break;

    case GameTransitionOut:

        assert(m_renderer);

        if (m_fadeValue >= FADE_SPEED_GAME)
        {
            m_fadeValue -= FADE_SPEED_GAME;
            m_isFading   = true;
        }
        else
        {
            m_fadeValue = 0.0;
            m_isFading  = false;
            m_state     = MenuTransitionIn;
        }

        m_renderer->setFadeValue(std::fmin(m_fadeValue, 1.0));

        break;

    case MenuTransitionIn:

        assert(m_renderer);

        if (m_fadeValue >= 1.0)
        {
            m_fadeValue = 1.0;
            m_isFading  = false;
            m_state     = Menu;

            // Re-init the track selection menu
            MTFH::MenuManager::instance().enterCurrentMenu();
        }
        else
        {
            m_fadeValue += FADE_SPEED_MENU;
            m_isFading   = true;
        }

        m_renderer->setFadeValue(m_fadeValue);

        break;

    default:
        break;
    }

    return true;
}

StateMachine::State StateMachine::state() const
{
    return m_state;
}

bool StateMachine::isFading() const
{
    return m_isFading;
}

void StateMachine::reset()
{
}
